<?php
date_default_timezone_set('UTC');

//define class autoloader
spl_autoload_register(function($className){
    $paths=array(
        '../classes/',
        '../classes/core/',
        '../classes/core/exception/',
        '../classes/core/authentication/',
        '../classes/core/interface/',
        '../classes/model/',
        '../classes/device/',
    );
    foreach($paths as $path){
        if(file_exists($path.$className.'.class.php')){
            include_once($path.$className.'.class.php');
        }
    }
});

//connect to the database
database::connect();
database::begin();

//set response type
$_REQUEST['responseType']='application/json'; // Later support xml if someone asks for it

$exceptionParameters=array();
$data=array();
try {

    $wwwroot=rtrim($_SERVER['DOCUMENT_ROOT'],'/');
    $uri=strtok($_SERVER['REQUEST_URI'],'?');
    $uri=rtrim($uri,'/');
    $uriArgs=explode('/', substr($uri,1));

    //get arguments from URI - ignore initial /device/
    array_shift($uriArgs);
    if(3===count($uriArgs)){
        $scope=$uriArgs[0];
        $action=$uriArgs[1];
        $apiKey=$uriArgs[2];
    } else if(2===count($uriArgs)){
        $scope=$uriArgs[0];
        $action=$uriArgs[1];
    } else if(1===count($uriArgs)){
        $scope=$uriArgs[0];
        echo json_encode(array('message'=>'Device access: '.$scope));
        exit;
    } else if(0===count($uriArgs)){
        echo json_encode(array('message'=>'This is the device access.'));
        exit;
    } else {
        throw new BadRequestException('POST to /device/DEVICE_SCOPE/DEVICE_ACTION');
    }

    //Do we have an API key, and is it valid for the scope (handler class)?
    //Accept API key in an X-IceBear-API-Key, Authorization, or Authorization: Bearer header,
    //or as third URI element, e.g., /device/SCOPE/METHOD/API_KEY
    if(!isset($apiKey)){
        $requestHeaders=apache_request_headers();
        if(isset($requestHeaders['X-IceBear-API-Key'])) {
            $apiKey = $requestHeaders['X-IceBear-API-Key'];
        } else if(isset($requestHeaders['Authorization'])){
            $header=$requestHeaders['Authorization'];
            $apiKey=trim(str_replace('Bearer', '', $header));
        } else {
            throw new AuthorizationRequiredException('No X-IceBear-API-Key or Authorization: Bearer header');
        }
    }

    //Did they send us a POST? It would be "cleaner" to insist on GET for read-only operations,
    //but there may be too much data to URL-encode, and the specs are vague on whether GET can
    //have a body, meaning that someone's proxy will probably eat it. So it's POST all the way.
    $method=$_SERVER['REQUEST_METHOD'];
    if("post"!==strtolower($method)){
        throw new BadRequestException("Use HTTP POST");
    }

    if(!apikey::validate($apiKey, $scope)){
        throw new AuthorizationRequiredException('API key not valid for scope '.$scope);
    }

    $logLevel=apikey::getLogLevel($apiKey);

    //Start logging
    Log::init($logLevel, 'DeviceAccess_'.$scope.'.log'); //TODO Log rotation
    Log::info("Request received, scope is $scope, action is $action");

    //attempt to start session
    session::init(new PhpSession());
    setScopeUserToSession($scope);

    //Convert post body JSON to an array
    $postBody=file_get_contents("php://input");
    $parameters=json_decode($postBody, true);
    if(null===$parameters){
        $parameters=json_decode(urldecode($postBody), true);
    }
    if(null===$parameters){
        throw new BadRequestException('Request body could not be decoded from JSON (bad character?), or exceeds recursion limit');
    }
    if(!$parameters){
        $parameters=array($parameters);
    }
    if(!$parameters){
        throw new BadRequestException('JSON in post body is invalid or exceeds recursion limit: '.$postBody);
    }
    $data=handleDeviceApiRequest($scope, $action, $parameters);

    //TODO if method starts with get, abort
    database::commit();

    respond($data,200);
} catch(Exception $e){
    database::abort();
    session::revertAdmin();
    $statusCode=$e->getCode();
    $trace=$e->getTrace();
    $prettytrace=array();
    $prettytrace[]=$e->getFile().' line '.$e->getLine();
    foreach($trace as $t){
        $traceline='';
        $line=(isset($t['line'])) ? $t['line'] : '--';
        $file=(isset($t['file'])) ? str_replace($_SERVER['DOCUMENT_ROOT'], '', (str_replace('\\','/',$t['file']))) : '--';
        if (array_key_exists('class',$t)){
            $traceline=sprintf("%s:%s %s::%s",
                $file,
                $line,
                $t['class'],
                $t['function']
            );
        } else {
            $traceline=sprintf("%s:%s %s",
                $file,
                $line,
                $t['function']
            );
        }
        $prettytrace[]=$traceline;
    }
    respond(array(
        'error'=>$e->getMessage(),
        'parameters'=>$exceptionParameters,
        'thrownat'=>$prettytrace[0],
        'trace'=>$prettytrace
    ),$statusCode);
    //TODO send stacktrace if admin
    //TODO Log error
} finally {
    Log::end();
    session::revokeAdmin();
}

/****************************************
 * Supporting functions below
 ***************************************/

/**
 * Returns the data to the client in the appropriate format, with the supplied HTTP response code.
 * @param array $data
 * @param int $responseCode
 */
function respond(array $data, int $responseCode){
    header('Access-Control-Allow-Origin: *');
    http_response_code($responseCode);
    header('Content-Type: application/json');
    $json=json_encode(sanitize($data));
    echo $json;
}

function sanitize($data){
    foreach($data as $k=>&$v){
        if(is_array($v)){
            $v=sanitize($v);
        } else {
            $v=htmlspecialchars($v);
        }
    }
    return $data;
}

/**
 * Calls a class method with the supplied arguments, throwing a BadRequestException if class or method does not exist.
 * @param string $className The class name
 * @param string $methodName The method name
 * @param array|null $args The arguments to pass to the method
 * @return mixed The results of the method call
 * @throws BadRequestException if the class or method does not exist
 */
function callOrThrow(string $className, string $methodName, $args=array()){
    if(!class_exists($className)){
        throw new BadRequestException('Class '.$className.' does not exist');
    }
    if(!method_exists($className, $methodName)){
        throw new BadRequestException('Class '.$className.': Method '.$methodName.' does not exist');
    }
    return forward_static_call_array(array($className, $methodName), $args);
}

/**
 * @param $scope
 * @param $action
 * @param $parameters
 * @return mixed
 * @throws BadRequestException
 * @throws ForbiddenException
 * @throws NotFoundException
 * @throws ServerException
 */
function handleDeviceApiRequest($scope, $action, $parameters){
    setScopeUserToSession($scope);
    if(!$scope || !$action) {
        throw new BadRequestException('Invalid URI. Call /device/SCOPE_NAME/METHOD_NAME');
    }
    return callOrThrow($scope, $action, array($parameters));
}

/**
 * @param string $scope
 * @throws BadRequestException
 * @throws ForbiddenException
 * @throws NotFoundException
 * @throws ServerException
 */
function setScopeUserToSession(string $scope){
    $username=callOrThrow($scope, 'getUsername');
    session::becomeAdmin();
    $user=user::getByName($username);
    if(!$user) {
        try {
            $password=bin2hex(random_bytes(10));
        } catch (Exception $exception){
            throw new ServerException('PHP random_bytes threw an exception');
        }
        $userFullName=callOrThrow($scope, 'getUserFullName');
        user::create(array(
            'name'=>$username,
            'fullname'=>$userFullName,
            'email'=>$username.'@bogus.example.com',
            'password'=>$password,
            'isactive'=>false
        ));
        $user=user::getByName($username);
    }
    $userId=$user['id'];
    session::set('user', $user);
    session::set('userId',$userId);
    session::revokeAdmin();
    session::refreshProjectPermissions();
}
